import re

class QueryCriteria(object):
    """
    A helper class to control query return-sets by filtering on basic properties and tags, as
    well as controlling which properties and child-objects to fill for returned objects.
    """

    def __init__(self):
        self._fields = []
        self._children = []
        self._filters = []
        self._tag_filters = []

    @property
    def fields(self):
        return self._fields

    @property
    def children(self):
        return self._children

    @property
    def filters(self):
        return self._filters

    @property
    def tag_filters(self):
        return self._tag_filters

    def select(self, fields):
        """
        Set which basic properties to fill for returned objects.

        :param fields: A list of basic properties to fill; e.g. ['id','description'].
        :return: A reference to this object so calls can be chained.
        """
        self._fields.extend([fields] if not isinstance(fields, list) else fields)
        return self

    def select_children(self, children):
        """
        Set which child objects to fill for returned objects.

        :param children: A list of child objects to fill; e.g. ['tags','hpc_jobs'].
        :return: A reference to this object so calls can be chained.
        """
        self._children.extend([children] if not isinstance(children, list) else children)
        return self

    def where(self, filters):
        """
        Set filter criteria for basic properties.

        For string filter values, '~' is used for the "like"-operator (i.e. string-contains).
        For numeric filter values, standard arithmetic operators are allowed.

        :param filters: A list of basic property filter-criteria; e.g. ['name~Test','state=Failed'].
        :return: A reference to this object so calls can be chained.
        """
        self._filters.extend([filters] if not isinstance(filters, list) else filters)
        return self

    def where_tag(self, tag_filters):
        """
        Set filter criteria for tags.

        For string filter values, '~' is used for the "like"-operator (i.e. string-contains).
        For numeric filter values, standard arithmetic operators are allowed.

        :param tag_filters: A list of tag filter-criteria; e.g. ['Replicate=3','DiseaseType~Malaria'].
        :return: A reference to this object so calls can be chained.
        """
        self._tag_filters.extend([tag_filters] if not isinstance(tag_filters, list) else tag_filters)
        return self

    def to_param_dict(self, ent_type):
        pd = {}

        if len(self._fields) > 0:
            pd['fields'] = ','.join(ent_type.py2rest(self._fields))

        if len(self._children) > 0:
            pd['children'] = ','.join(ent_type.py2rest(self._children))

        if len(self._filters) > 0:
            tups = [ (f, re.search('\W', f)) for f in self._filters ]

            keys = [ f[:r.start()] if r else f for f, r in tups ]
            vals = [ f[r.start():] if r else f for f, r in tups ]

            mod_filters = map(lambda x, y: x + y, ent_type.py2rest(keys), vals)

            pd['filters'] = ','.join(mod_filters)

        if len(self._tag_filters) > 0:
            pd['tagfilters'] = ','.join(self._tag_filters)

        return pd
